#
# Copyright 2017-2020 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#
# Refer to the README and COPYING files for full details of the license
#
import contextlib

from ovirtsdk4 import types

from ovirtlib import error
from ovirtlib import netlib
from ovirtlib import storagelib
from ovirtlib import syncutil
from ovirtlib.sdkentity import SDKRootEntity


class DataCenter(SDKRootEntity):

    @property
    def name(self):
        return self.get_sdk_type().name

    @property
    def status(self):
        return self.get_sdk_type().status

    def attach_storage_domain(self, sd):
        sds_service = self._service.storage_domains_service()
        sds_service.add(sd.get_sdk_type())

    def deactivate_storage_domain(self, sd):
        self._sd_service(sd).deactivate()

    def deactivate_storage_domain_sync(self, sd):
        syncutil.sync(
            exec_func=self.deactivate_storage_domain,
            exec_func_args=(sd,),
            error_criteria=error.sd_deactivation_error_not_due_to_busy)

    def list_qos(self):
        qos_entities = []
        for qos in self._service.qoss_service().list():
            qos_entity = netlib.QoS(self)
            qos_entity.import_by_name(name=qos.name)
            qos_entities.append(qos_entity)
        return qos_entities

    def remove_qos(self, qos_names):
        for qos in self.list_qos():
            if qos.name in qos_names:
                self._service.qoss_service().qos_service(qos.id).remove()

    def wait_for_up_status(self):
        self._wait_for_status(types.DataCenterStatus.UP)

    def wait_for_sd_active_status(self, sd):
        self._wait_for_sd_status(sd, storagelib.StorageDomainStatus.ACTIVE)

    def wait_for_sd_maintenance_status(self, sd):
        self._wait_for_sd_status(sd,
                                 storagelib.StorageDomainStatus.MAINTENANCE)

    def create(self, dc_name):
        sdk_type = types.DataCenter(name=dc_name)
        self._create_sdk_entity(sdk_type)

    def get_mgmt_network(self):
        ovirtmgmt = netlib.Network(self)
        ovirtmgmt.import_by_name(netlib.OVIRTMGMT)
        return ovirtmgmt

    def _get_parent_service(self, system):
        return system.data_centers_service

    def _wait_for_sd_status(self, sd, status):
        sd_service = self._sd_service(sd)
        syncutil.sync(exec_func=lambda: sd_service.get().status,
                      exec_func_args=(),
                      success_criteria=lambda s: s == status)

    def _sd_service(self, sd):
        return self._service.storage_domains_service().service(sd.id)

    def _wait_for_status(self, status):
        syncutil.sync(exec_func=lambda: self.status,
                      exec_func_args=(),
                      success_criteria=lambda s: s == status,
                      timeout=60 * 5)

    @staticmethod
    def iterate(system, search=None):
        """Iterate over `system`'s data centers

        If `search` is None, all data centers are iterated. Otherwise,
        `search` should be a search query string understood by oVirt.
        Cf. http://ovirt.github.io/ovirt-engine-api-model/master/#_searching
        """
        for sdk_obj in system.data_centers_service.list(search=search):
            dc = DataCenter(system)
            dc.import_by_id(sdk_obj.id)
            yield dc


@contextlib.contextmanager
def attached_storage_domain(data_center, storage_domain):
    data_center.attach_storage_domain(storage_domain)
    try:
        # assuming that even if the storage does not become active,
        # deactivate will not fail:
        data_center.wait_for_sd_active_status(storage_domain)
        yield storage_domain
    finally:
        data_center.deactivate_storage_domain_sync(storage_domain)
        data_center.wait_for_sd_maintenance_status(storage_domain)
